import numpy as np
from scipy.special import gammaln

from first_diff import opDx
from first_diff import opDtx
from RLTV_utils import compute_grad_mag
from RLTV_utils import plot_hist_to_screen, plot_grad_hist_to_screen
from RLTV_utils import plot_grad_mag_seq, plot_grad_hist_to_file
from RLTV_utils import compute_gradient_std_dev, plot_hist_to_file
from RLTV_utils import compute_pct_under_threshold



def run_RLTV(nIters, x0, b, Ax, Atx, lam, use_reg_term, grad_mag_threshold, \
    x_gt, grad_mag_dirname=None, grad_hist_dirname=None, update_rule_dirname=None):

    x = x0
    sum_log_b_factorial = np.sum(gammaln(b + 1.0))
    ones_img = np.ones_like(x)
    At_ones = Atx(ones_img)

    # new arrays for result statistics
    log_likelihood = np.zeros(nIters, float)
    mse = np.zeros(nIters, float)
    dxyl1norm = np.zeros(nIters, float)
    gradient_std_dev = np.zeros(nIters, float)
    pct_grad_mag_threshold = np.zeros(nIters, float)

    # initialize terms needed for x update before entering loop
    A_of_x = Ax(x)
    A_of_x = np.where(np.isnan(A_of_x), 0, A_of_x)
    A_of_x = np.where(A_of_x <= 0, 1e-14, A_of_x)
    Dx = opDx(x,axis=-1)
    Dy = opDx(x,axis=-2)
    abs_Dx = np.fabs(Dx)
    abs_Dx_nz = np.where(abs_Dx == 0.0, 1.0, abs_Dx)
    abs_Dy = np.fabs(Dy)
    abs_Dy_nz = np.where(abs_Dy == 0.0, 1.0, abs_Dy)
    Dx_divide_abs_Dx = Dx / abs_Dx_nz
    Dy_divide_abs_Dy = Dy / abs_Dy_nz
    reg_Dx_term_A = Dx_divide_abs_Dx
    reg_Dy_term_A = Dy_divide_abs_Dy
    reg_Dx_term_B = opDtx(Dx_divide_abs_Dx, axis=-1)
    reg_Dy_term_B = opDtx(Dy_divide_abs_Dy, axis=-2)

    for i in range(0, nIters):
        if use_reg_term == 'RL+TV_B':
            # "correct"
            reg_term = lam * (reg_Dx_term_B + reg_Dy_term_B)
            x = (Atx(b/A_of_x) / (At_ones + reg_term)) * x
            if update_rule_dirname != None:
                plot_hist_to_file(At_ones + reg_term, update_rule_dirname, \
                    "Rule B Denom " + use_reg_term + " \u03BB=" + str(lam), \
                    "Rule_B_Denom_" + use_reg_term + "_\u03BB=" + str(lam), i)
        elif use_reg_term == 'RL+TV_A':
            # "incorrect"
            reg_term = lam * (reg_Dx_term_A + reg_Dy_term_A)
            x = (Atx(b/A_of_x) / (At_ones - reg_term)) * x
            if update_rule_dirname != None:
                plot_hist_to_file(At_ones - reg_term, update_rule_dirname, \
                    "Rule A Denom " + use_reg_term + " \u03BB=" + str(lam), \
                    "Rule_A_Denom_" + use_reg_term + "_\u03BB=" + str(lam), i)
        else:
            print("Please select a regularization term")

        x = np.clip(x, a_min=0.0, a_max=None) # make sure non-negative

        # recompute terms needed at top of loop based on updated x
        A_of_x = Ax(x)
        A_of_x = np.where(np.isnan(A_of_x), 0, A_of_x)
        A_of_x = np.where(A_of_x <= 0.0, 1e-14, A_of_x)
        Dx = opDx(x, axis=-1)
        Dy = opDx(x, axis=-2)
        abs_Dx = np.fabs(Dx)
        abs_Dx_nz = np.where(abs_Dx == 0.0, 1.0, abs_Dx)
        abs_Dy = np.fabs(Dy)
        abs_Dy_nz = np.where(abs_Dy == 0.0, 1.0, abs_Dy)
        Dx_divide_abs_Dx = Dx / abs_Dx_nz
        Dy_divide_abs_Dy = Dy / abs_Dy_nz
        reg_Dx_term_A = Dx_divide_abs_Dx
        reg_Dy_term_A = Dy_divide_abs_Dy
        reg_Dx_term_B = opDtx(Dx_divide_abs_Dx, axis=-1)
        reg_Dy_term_B = opDtx(Dy_divide_abs_Dy, axis=-2)

        sum_Ax = np.sum(A_of_x)
        log_Ax_dotprod_b = np.sum(np.log(A_of_x) * b)
        DxyL1norm = np.sum(abs_Dx) + np.sum(abs_Dy)
        log_likelihood[i] = log_Ax_dotprod_b - sum_Ax - \
            sum_log_b_factorial - lam * DxyL1norm
        mse[i] = np.mean((x_gt - x) ** 2)
        dxyl1norm[i] = DxyL1norm
        gradient_std_dev[i] = compute_gradient_std_dev(Dx, Dy)
        x_grad_mag = compute_grad_mag(x)
        pct_grad_mag_threshold[i] = \
            compute_pct_under_threshold(x_grad_mag, grad_mag_threshold)

        if grad_mag_dirname != None:
            plot_grad_mag_seq(x_grad_mag, grad_mag_dirname, \
                use_reg_term, lam, i)

        if grad_hist_dirname != None:
            plot_grad_hist_to_file(x, grad_hist_dirname, \
                "Gradient Histogram " + use_reg_term + " \u03BB=" + str(lam), \
                "Gradient_Histogram_" + use_reg_term + "_\u03BB=" + str(lam), i)

        # uncomment to generate gradient magnitude histograms per iteration
        # plot_hist_to_screen(x_grad_mag)

        # uncomment to generate gradient histograms per iteration
        # plot_grad_hist_to_screen(x)

    return [x, log_likelihood, mse, dxyl1norm, gradient_std_dev, \
        pct_grad_mag_threshold]

