import numpy as np
from numpy.fft import fft2, ifft2
import matplotlib.pyplot as plt
import skimage.io as io
from numpy.random import poisson
from numpy.random import random_sample
from pypher.pypher import psf2otf
from pathlib import Path
from textwrap import wrap
from scipy.special import gammaln
import os
import sys

from fspecial import fspecial_gaussian_2d
from first_diff import opDx
from first_diff import opDtx
from RLTV_utils import compute_grad_mag, plot_grad_mag
from RLTV_utils import plot_hist_to_screen, plot_grad_hist_to_screen
from RLTV_utils import plot_grad_mag_seq, plot_grad_hist_to_file
from RLTV_utils import compute_gradient_std_dev, plot_hist_to_file
from RLTV_utils import compute_pct_under_threshold, check_create_dir
from RLTV_utils import compute_ground_truth_stats
from RLTV import run_RLTV

def compute_psnr(x_true, x_recon, n_photons):
    mse = np.mean((x_true - x_recon) ** 2)
    psnr = 10 * np.log10(n_photons ** 2 / mse)
    return psnr

# Specify below in param_list for each set of parameters to jointly plot:
# 1) the regularization term for the RL+TV prior x update rule, either
#    RL+TV_A or RL+TV_B
# 2) the value of lambda
#
# as a tuple.  param_list may hold one or more tuples.
#
# RL+TV_A is the x update rule from the original EE367 Lecture 10 notes, Eq.27
# downloaded on 19 February 2021
#
# RL+TV_B is the x update rule from the updated EE367 Lecture 10 notes, Eq. 28
# downloaded on 22 February 2021
#
# lambda is typically between 0.01 and 0.05
#
# Example:
# param_list = [("RL+TV_A", 0.04), ("RL+TV_B, 0.02)]
# 
# Also, a plot_id string may be provided
# Example:
# plot_id = "joint_plot_1"
#
# Also, set nIters as appropriate

param_list = [("RL+TV_A", 0.04), ("RL+TV_B", 0.02)]
plot_id = "joint_plot_1"
nIters = 100

filenum = 1

if __name__ == "__main__":
    for i, arg in enumerate(sys.argv):
        if arg.isnumeric():
            filenum = int(arg)
            if (filenum <= 0) or (filenum > 7):
                filenum = 1

print("Using image file " + str(filenum))

image_names = [("FluorescentCells.jpg", "cells"),
               ("hubble_deep_field_1.png", "hubble"),
               ("art_rby_1.png", "art"),
               ("trv-res_1.png", "testpat"),
               ("coffee.png", "coffee"),
               ("motorcycle_right.png", "motorcycle"),
               ("bluecells.png", "cells_blue")]

img_pathname = 'test_images/' + image_names[filenum - 1][0]
img_str = image_names[filenum - 1][1]
print("Using " + img_pathname)

outdir = "./" + "RL_TV_test_joint_plots"
check_create_dir(outdir)

img = io.imread(img_pathname).astype(np.float32)/255
height, width = img.shape[0], img.shape[1]
if img.shape[2] > 3:
    # image has an alpha channel
    # in the set of test images, all images are opaque (alpha = 1.0)
    # therefore, just ignore alpha
    img = img[:,:,0:3]
x_gt = img.transpose(2, 0, 1)

# PSF and OTF
psf = fspecial_gaussian_2d((15, 15), 1.5)
otf = psf2otf(psf, (height, width))

# Number of photons
n_photons = 100
x_gt = x_gt * n_photons

# Function handles for 2D convolution
Ax = lambda x: np.real(ifft2(fft2(x) * otf))
Atx = lambda x: np.real(ifft2(fft2(x) * np.conj(otf)))

# b for all parameter combinations
# for a given image file, attempt to use same b for all runs
# of the program by storing the blurred, noisy values in a file
fname = 'test_images/' + "/" + img_str + "_b.dat"
if os.access(fname, os.R_OK):
    f = open(fname, "rb")
    b = np.load(f)
    f.close()
else:
    b_blur = Ax(x_gt)
    # Make sure there are no negative values
    b_blur[b_blur < 0] = 0

    # Add poisson noise
    b = poisson(b_blur)
    f = open(fname, "wb")
    np.save(f,b)
    f.close()

# compute gradient magnitude of b
b_grad_mag = compute_grad_mag(b)

# define a threshold for per pixel gradient magnitude
# sparser gradient images will have a higher percentage of pixels
# below the threshold
grad_mag_threshold = 0.5 * np.std(b_grad_mag)

# initial x for all parameter combinations
# for a given image file, attempt to use same initial x for all runs
# of the program by storing the random values in a file
fname = 'test_images/' + "/" + img_str + "_x_init.dat"
if os.access(fname, os.R_OK):
    f = open(fname, "rb")
    x0 = np.load(f)
    f.close()
else:
    x0 = np.random.rand(*x_gt.shape)
    f = open(fname, "wb")
    np.save(f,x0)
    f.close()

log_likelihoods = {}
mses = {}
dxyl1norms = {}
gradient_std_devs = {}
pct_grad_mag_thresholds = {}

for params in param_list:

    use_reg_term = params[0]
    lam = params[1]

    results = run_RLTV(nIters, x0, b, Ax, Atx, lam, use_reg_term, \
        grad_mag_threshold, x_gt)

    x = results[0]
    log_likelihoods[use_reg_term, lam] = results[1]
    mses[use_reg_term, lam] = results[2]
    dxyl1norms[use_reg_term, lam] = results[3]
    gradient_std_devs[use_reg_term, lam] = results[4]
    pct_grad_mag_thresholds[use_reg_term, lam] = results[5]

# plot log likelihoods vs iterations for all parameter combinations
plt.figure()
for params in param_list:
    use_reg_term = params[0]
    lam = params[1]
    plt.plot(log_likelihoods[use_reg_term, lam], \
        label=use_reg_term + ', \u03BB=' + str(lam))
plt.legend()
plt.title("Log Likelihood")
plt.xlabel('Iteration')
plt.ylabel('Log Likelihood')
fname = outdir + "/" + plot_id + "_log_likelihood.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# plot MSEs vs iterations for all parameter combinations
plt.figure()
for params in param_list:
    use_reg_term = params[0]
    lam = params[1]
    plt.plot(mses[use_reg_term, lam], \
        label=use_reg_term + ', \u03BB=' + str(lam))
plt.legend()
plt.title("MSE")
plt.xlabel('Iteration')
plt.ylabel('MSE')
fname = outdir + "/" + plot_id + "_mse.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# compute ground truth stats to include in plots below
results = compute_ground_truth_stats(x_gt, grad_mag_threshold)
x_gt_dxyl1norm = np.zeros(nIters, float)
x_gt_dxyl1norm[:] = results[0]
x_gt_gradient_std_dev = np.zeros(nIters, float)
x_gt_gradient_std_dev[:] = results[1]
x_gt_pct_grad_mag_threshold = np.zeros(nIters, float)
x_gt_pct_grad_mag_threshold[:] = results[2]

# plot dxyl1norms vs iterations for all parameter combinations
plt.figure()
for params in param_list:
    use_reg_term = params[0]
    lam = params[1]
    plt.plot(dxyl1norms[use_reg_term, lam], \
        label=use_reg_term + ', \u03BB=' + str(lam))
plt.plot(x_gt_dxyl1norm, label="ground truth", \
    linestyle=(0, (2, 3)))
plt.legend()
plt.title("Dx+Dy L1 Norm")
plt.xlabel('Iteration')
plt.ylabel('L1 Norm')
fname = outdir + "/" + plot_id + "_dxyl1norm.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# plot gradient std deviations vs iterations for all parameter combinations
plt.figure()
for params in param_list:
    use_reg_term = params[0]
    lam = params[1]
    plt.plot(gradient_std_devs[use_reg_term, lam], \
        label=use_reg_term + ', \u03BB=' + str(lam))
plt.plot(x_gt_gradient_std_dev, label="ground truth", \
    linestyle=(0, (2, 3)))
plt.legend()
plt.title("Gradient Std Dev")
plt.xlabel('Iteration')
plt.ylabel('Std Dev')
fname = outdir + "/" + plot_id + "_grad_std_dev.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()

# plot percent under threshold vs iterations for all parameter combinations
plt.figure()
for params in param_list:
    use_reg_term = params[0]
    lam = params[1]
    plt.plot(pct_grad_mag_thresholds[use_reg_term, lam], \
        label=use_reg_term + ', \u03BB=' + str(lam))
plt.plot(x_gt_pct_grad_mag_threshold, label="ground truth", \
    linestyle=(0, (2, 3)))
plt.legend()
plt.title("Percent Gradient Magnitude Under Threshold")
plt.xlabel('Iteration')
plt.ylabel('Percent')
fname = outdir + "/" + plot_id + "_pct_grad_mag_threshold.png"
plt.savefig(fname, bbox_inches='tight')
plt.close()





